#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/slab.h>

#define KTOP_TIMES1_VALID (1<<0)
#define KTOP_TIMES2_VALID (1<<1)
#define KTOP_ALL_TIMES_VALID ((KTOP_TIMES1_VALID)|(KTOP_TIMES2_VALID))

/* TCMEM_SIZE sets the maximum number of threads to track */

#define TCMEM_SIZE 200
#define PIDS_TO_PRINT 5

struct ktop_task_cache {
	int pid;
	unsigned long rss;
	unsigned long stime1, utime1, stime2, utime2;
	unsigned char rt_priority;
	unsigned char policy;
	unsigned char age;
	char valid;
	char cmd[16];
	char state[4];
	struct ktop_task_cache* next;
};

static struct ktop_task_cache **tchash;

static struct ktop_task_cache *tcmem; /* Handles at most TCMEM_SIZE pids at once */
static int tcmem_offset=0;
static int ktop_escaping=0;

static unsigned long jiffies_in_sample = 0;
static char* rtpolicy_string[]={"NONE", "FIFO", " RR "};
static unsigned long long last_run_jiffies=0;
static int runs_todo = 0;
static struct timer_list ktop_timer;
static spinlock_t ktop_spinlock=SPIN_LOCK_UNLOCKED;

struct ktop_task_cache* ktop_alloc(void) {
	int i;
	while(1) {
		for(i=tcmem_offset; i < TCMEM_SIZE; i++) {
			if(tcmem[i].age == 0) {
				tcmem_offset=i;
				return &tcmem[i];
			}
		}
		if(tcmem_offset == 0) {
			/* Eek! */
			printk("KTOP: Ran out of TCMEM SPACE!!!\n");
			return NULL;
		}
		tcmem_offset=0;
	}
}

void ktop_dealloc (struct ktop_task_cache* tcc) {
	struct ktop_task_cache* tccp;
	/* Case entry is at top of hash */
	if(tchash[pid_hashfn(tcc->pid)] == tcc) {
		tchash[pid_hashfn(tcc->pid)]=tcc->next;
	} else {
		/* Uncommon case, not at top of hash
		   walk link list to find */
		for(tccp=tchash[pid_hashfn(tcc->pid)];tccp->next != NULL && tccp->next != tcc; tccp=tccp->next);
		tccp->next=tcc->next;
	}
	tcc->next = NULL;
}

static struct ktop_task_cache* ktop_find_by_pid(int pid) {
	struct ktop_task_cache *ttc;
	for (ttc=tchash[pid_hashfn(pid)]; ttc != NULL && ttc->pid != pid; ttc=ttc->next);
	if(ttc == NULL) {
		ttc= ktop_alloc();
		if(ttc == NULL)
			return NULL;
		ttc->pid=pid;
		ttc->valid=0;
		ttc->next=tchash[pid_hashfn(pid)];
		tchash[pid_hashfn(pid)]=ttc;
	}
	return ttc;
}

void ktop_update_task(struct task_struct* ts) {
	struct ktop_task_cache* ttc= ktop_find_by_pid(ts->pid);
	if(ttc == NULL) {
		printk("Unable to updated task with PID %d\n", ts->pid);
		return;
	}
	if(ttc->valid) {
		/* This executes the second pass through this
		   loop for any given task.  This is because
		   the TIME2 is only valid on the the second pass */
		ttc->stime1=ttc->stime2;
		ttc->utime1=ttc->utime2;
		/* Set TIMES1 valid also, now we can report the
		   CPU usage */
		ttc->valid |= KTOP_TIMES1_VALID;
	} else {
		memcpy(ttc->cmd, ts->comm, 16);
		ttc->cmd[15]='\0';
		ttc->state[1]=' ';
		ttc->state[2]=' ';
		ttc->state[3]='\0';
	}
	ttc->stime2=ts->times.tms_stime;
	ttc->utime2=ts->times.tms_utime;
	ttc->rt_priority=ts->rt_priority;
	ttc->policy=ts->policy;
	ttc->rss=ts->mm?ts->mm->rss:0;
	/* KTOP_TIMES2_VALID gets set first */
	ttc->valid |= KTOP_TIMES2_VALID;
	ttc->age=2;
	switch(ts->state) {
	case TASK_RUNNING:
		ttc->state[0]='R';
		break;
	case TASK_INTERRUPTIBLE:
		ttc->state[0]='S';
		break;
	case TASK_UNINTERRUPTIBLE:
		ttc->state[0]='D';
		break;
	case TASK_ZOMBIE:
		ttc->state[0]='Z';
		break;
	case TASK_STOPPED:
		ttc->state[0]='T';
		break;
	default:
		ttc->state[0]='U';
		break;
	}

}

/* Nothing in here can schedule!!! */
void ktop_update_tasks(void) {
	int i;
	struct task_struct* ts;
	for(i=0; i < TCMEM_SIZE; i++) {
		if(tcmem[i].age > 0) {
			tcmem[i].age--;
			if(tcmem[i].age == 0) {
				/* Clean up */
				ktop_dealloc(&tcmem[i]);
			}
		}
	}
	for(i=0; i < PIDHASH_SZ; i++) {
		for(ts = pidhash[i]; ts != NULL; ts=ts->pidhash_next) {
			ktop_update_task(ts);
		}
	}
}

void ktop_print_cache(void) {
	int pids[PIDS_TO_PRINT][2];
	int i;
	if( jiffies_in_sample == 0 )
		return;
	memset(pids, 0, sizeof(unsigned long) * PIDS_TO_PRINT * 2);
	for(i=0; i<TCMEM_SIZE; i++) {
                /*
		  Funky sort..., most deltas will be zero on
		  our normal case of about 100 items, so it
		  should be reasonably fast.
		*/
		if(tcmem[i].age != 2)
			continue;
		if((tcmem[i].valid & KTOP_ALL_TIMES_VALID)
		   == KTOP_ALL_TIMES_VALID) {
			int pos=PIDS_TO_PRINT-1;
			int delta=tcmem[i].stime2-tcmem[i].stime1 +
				tcmem[i].utime2-tcmem[i].utime1;
			if(delta > 999)
				delta=999;
			if(delta < 0)
				delta = 0;
			if(delta > pids[pos][0]) {
				while(pos--){
					if(delta <= pids[pos][0]) {
						break;
					} else {
						pids[pos+1][0]=pids[pos][0];
						pids[pos+1][1]=pids[pos][1];
					}
				}
				pids[pos + 1][0]=delta;
				pids[pos + 1][1]=i;
			}
		}
	}
	if(ktop_escaping == 0)
		printk(KERN_INFO);
	printk("%s PID  STATUS   RSS POL  PRI %%CPU COMMAND %s\n",
	       ktop_escaping == 1? "\n\e[7m": "", ktop_escaping == 1?"\e[0m":"");

	for(i=0; i<PIDS_TO_PRINT; i++) {
		if(pids[i][0] == 0) {
			return;
		} else {
			struct ktop_task_cache* ttc=&tcmem[pids[i][1]];
			char rss_buf[8];
			int pcpu;
			if(ttc->rss > 10 * 1024) 
				sprintf(rss_buf, "%6ldM", ttc->rss/1024);	
			else
				sprintf(rss_buf, "%7ld", ttc->rss);
			if(ttc->policy > 2)
				ttc->policy=0;
			
			pcpu=(pids[i][0] * 1000) / jiffies_in_sample;
			if(ktop_escaping == 0)
				printk(KERN_INFO);
			printk("%5d %s  %s %-4s %3d %2d.%d %s\n",
			       ttc->pid, ttc->state, rss_buf,
			       rtpolicy_string[ttc->policy], ttc->rt_priority,
			       pcpu/10, pcpu%10, ttc->cmd);
		}
	}
}

void ktop_timer_handler (unsigned long ptr) {
	ktop_update_tasks();
	if(last_run_jiffies != 0)
		jiffies_in_sample=jiffies-last_run_jiffies;
	last_run_jiffies=jiffies;
	ktop_print_cache();
	if(runs_todo-- > 0) {
		mod_timer(&ktop_timer, jiffies + HZ);
	} else {
		kfree(tchash);
		tchash=NULL;
		kfree(tcmem);
		tcmem=NULL;
	}
}

/* *must* be executed in process context */
void ktop_start(int seconds, int escaping) {
	spin_lock(&ktop_spinlock);
	if( runs_todo <= 0 && seconds > 0) {
		/* 2 pages */
		runs_todo=seconds;
		spin_unlock(&ktop_spinlock);
		tchash= kmalloc(sizeof(struct ktop_task_cache*) * PIDHASH_SZ, GFP_KERNEL);
		if( tchash == NULL ) {
			printk("top_kernel_start: Unable to allocate PIDHASH\n");
			return;
		}
		memset(tchash, 0, sizeof(struct ktop_task_cache*) * PIDHASH_SZ);
		/* 3 pages */
		tcmem= kmalloc(sizeof(struct ktop_task_cache) * TCMEM_SIZE, GFP_KERNEL);
		if( tcmem == NULL ) {
			printk("top_kernel_start: Unable to allocate TCMEM\n");
			kfree(tchash);
			return;
		}
		memset(tcmem, 0, sizeof(struct ktop_task_cache) * TCMEM_SIZE);

		init_timer (&ktop_timer);
		ktop_timer.function=ktop_timer_handler;
		ktop_timer.expires = jiffies + HZ;
		ktop_escaping=escaping;
		last_run_jiffies=0;
		add_timer (&ktop_timer);
	} else {
		spin_unlock(&ktop_spinlock);	     
		printk("Kernel CPU Usage already running!\n");
	}
}
