/*
   Common database defines for the kbuild pre-processors.

   (c) Keith Owens <kaos@ocs.com.au> May 2002.
   Released under GPL version 2, (you know the rest).
 */

#ifndef _PP_DB_H
#define _PP_DB_H

#include <assert.h>
#include <sys/types.h>
#include <stddef.h>

/* ---- Record formats ---- */

/* Any change to record formats must bump this version number which will erase
 * the old database and force a complete kernel rebuild.  There is no support
 * for migrating a kbuild database from one version to another, rebuild the
 * database from scratch.
 */
#define DB_VERSION	9

typedef struct ppdb				PPDB;
typedef struct db_token				DB_TOKEN;

/* See declare_db_list() below for most of the _LIST declarations */
typedef struct db_rec_filename_1		DB_REC_FILENAME_1;
typedef struct db_rec_filename_list		DB_REC_FILENAME_LIST;
typedef struct db_rec_text			DB_REC_TEXT;
typedef struct db_rec_directory			DB_REC_DIRECTORY;
typedef struct db_rec_target			DB_REC_TARGET;
typedef struct db_rec_source			DB_REC_SOURCE;

typedef unsigned char				DB_TYPE;

#define extra_aflags	acld_flags[0]
#define extra_cflags	acld_flags[1]
#define extra_ldflags	acld_flags[2]
#define strip_aflags	acld_flags[3]
#define strip_cflags	acld_flags[4]
#define strip_ldflags	acld_flags[5]

/* Hide the underlying database engine from the pre-processing programs.
 * The data is implemented as a kbuild specific database, but most of the code
 * does not need to know that.
 */

struct ppdb;

/* Pointer to another database record */
struct db_token {
	DB_TYPE type;
	unsigned int recno: 24;
};

/* Information about database keys and records */
typedef struct {
	void *dptr;
	unsigned int dsize;
} datum;

typedef struct {
	datum key;
	datum val;
} kvpair;


#define DB_TOKEN_NULL		{ .type = '\0', .recno = 0 }
#define DB_MAX_RECNO (1<<24)

#define DB_IS_NULL_TOKEN(x) ((x).type == '\0' && (x).recno == 0)

#define DB_RECTYPE_BASE_LIST		'b'
#define DB_RECTYPE_FILENAME_LIST	'/'
#define DB_RECTYPE_DIRECTORY		'd'
#define DB_RECTYPE_BASE_SELECT_LIST	'B'
#define DB_RECTYPE_SELECT_LIST		'S'
#define DB_RECTYPE_TARGET		't'
#define DB_RECTYPE_SOURCE		's'
#define DB_RECTYPE_CHILD_LIST		'h'
#define DB_RECTYPE_TARGET_LIST		'g'
#define DB_RECTYPE_CCD_LIST		'C'
#define DB_RECTYPE_TEXT			'T'

#define DB_KEY_BASE_LIST { .type = DB_RECTYPE_BASE_LIST, .recno = 1 }

#define declare_db_list(U, l) \
	typedef struct db_rec_##l##_1 DB_REC_##U##_1; \
	typedef struct db_rec_##l##_list DB_REC_##U##_LIST; \
	extern void new_db_##l##_list(PPDB * db, DB_TOKEN * list_p); \
	extern DB_REC_##U##_LIST *read_db_##l##_list(PPDB * db, DB_TOKEN list); \
	extern DB_REC_##U##_LIST *copy_db_##l##_list(PPDB * db, DB_TOKEN list); \
	extern void empty_db_##l##_list(PPDB * db, DB_TOKEN list); \
	extern DB_REC_##U##_LIST *add_db_##l##_list(PPDB * db, DB_TOKEN list, const DB_REC_##U##_1 *add, DB_REC_##U##_1 **added);

declare_db_list(BASE, base)
declare_db_list(BASE_SELECT, base_select)
declare_db_list(SELECT, select)
declare_db_list(CHILD, child)
declare_db_list(TARGET, target)
declare_db_list(CCD, ccd)

/* See declare_db_list_size() below for most of the _LIST size macros */
#define DB_SIZE_FILENAME_1(rec)		(sizeof(*(rec)))
#define DB_SIZE_FILENAME_LIST(rec)	(sizeof(*(rec))+((rec)->count)*sizeof((rec)->list[0]))
/* No size macro for _TEXT */
#define DB_SIZE_DIRECTORY(rec)		(sizeof(*(rec)))
#define DB_SIZE_TARGET(rec)		(sizeof(*(rec)))
#define DB_SIZE_SOURCE(rec)		(sizeof(*(rec)))

/* text records */
struct db_rec_text {
	char text[0];
};

/* One base record that lists all the known bases */
struct db_rec_base_1 {
	struct {
		unsigned int has_modules: 1;
		unsigned int has_arch_head: 1;
		unsigned int is_host: 1;
	} flags;
	DB_TOKEN entry;			/* name of the base */
	DB_TOKEN topdir;		/* the top directory the base is used in */
};

struct db_rec_base_list {
	int count;			/* count of bases defined */
	DB_REC_BASE_1 list[0];
};

/* directory records */
struct db_rec_directory {
	DB_TOKEN relname;
	DB_TOKEN parent;		/* -> parent directory */
	struct {
		unsigned int create_objdir: 1;
		unsigned int explicit: 1;
		unsigned int walked: 1;		/* directory walked for printing */
		unsigned int makefile: 1;	/* contains a Makefile.in that is being used */
	} flags;
	DB_TOKEN acld_flags[6];
	DB_TOKEN children;		/* files and immediate sub-directories */
	DB_TOKEN base_select;		/* -> select lists in this directory for all bases */
};

/* One variable sized base_select record per directory */
struct db_rec_base_select_1 {
	DB_TOKEN entry;			/* -> select list */
};

struct db_rec_base_select_list {
	int count;			/* count of bases defined */
	DB_REC_BASE_SELECT_1 list[0];
};

/* target records */
struct db_rec_target {
	DB_TOKEN relname;
	DB_TOKEN parent;			/* -> parent directory */
	time_t mtime;				/* last modified time */
	struct {
		unsigned int nodepend: 1;
		unsigned int setup: 1;
		unsigned int expsyms: 1;
		unsigned int archive: 1;
		unsigned int archive_member: 1;
		unsigned int object: 1;
		unsigned int assembler: 1;
		unsigned int gen_i: 1;		/* generate a *.i file */
		unsigned int gen_s: 1;		/* generate a *.s file */
		unsigned int makecmdgoal: 1;	/* specified on command line */
		unsigned int makefile: 1;	/* specified in makefile */
		unsigned int commands: 1;	/* a makefile contains explicit commands for this target */
		unsigned int commands_gen: 1;	/* commands have been generated for this target */
		unsigned int arch_head: 1;	/* This target is part of $(arch_head) */
		unsigned int sely_summary: 1;	/* summary over sely for all bases */
		unsigned int selm_summary: 1;	/* summary over selm for all bases */
		unsigned int host_summary: 1;	/* summary of host flags over all selected bases */
		unsigned int issrcfile: 1;	/* referenced by srcfile/srcfile_base */
		unsigned int uptodate: 1;	/* target is up to date */
		unsigned int multi_linked: 1;	/* linked into multiple objects */
		unsigned int warned_multi_linked: 1;	/* warned about multiple linkage */
		unsigned int walked: 1;		/* target walked for printing */
		unsigned int ignore: 1;		/* dead entry, ignore it */
		unsigned int symlink: 1;	/* filename is a symlink */
	} flags;
	DB_TOKEN source;			/* source file this target is generated from */
	DB_TOKEN dependency;			/* files this target depends on */
	DB_TOKEN objlink;
	DB_TOKEN command;			/* the last command used to successfully build the target */
	DB_TOKEN acld_flags[6];
	DB_TOKEN link;				/* the conglomerate or archive this target links into, if any */
};

/* source records */
struct db_rec_source {
	DB_TOKEN relname;
	DB_TOKEN parent;		/* -> parent directory */
	int index;			/* which tree the source is in */
	time_t mtime;			/* last modified time */
	struct {
		unsigned int uptodate: 1;
		unsigned int config_extracted: 1;
		unsigned int shadow_source: 1;
		unsigned int setup: 1;
	} flags;
	DB_TOKEN component;		/* files this source is built from, if any */
	DB_TOKEN new_component;		/* new version of above list to detect changes */
	DB_TOKEN config;		/* configs this source depends on, if any */
};

/* Filename records are keyed by filename relative to the source/object root,
 * they are used for fast name lookup.
 */
struct db_rec_filename_1 {
	int index;			/* index of this entry, replicated to simplify filename processing */
	DB_TOKEN entry;			/* -> text, target, source or directory record */
};

struct db_rec_filename_list {
	int count;
	struct db_rec_filename_1 list[0];
};

/* List of selected entries and their select status */
struct db_rec_select_1 {
	struct {
		unsigned int sely: 1;
		unsigned int selm: 1;
		unsigned int sels: 1;
		unsigned int objlinked: 1;	/* selected via objlink list */
		unsigned int walked: 1;		/* walked via select chain */
	} flags;
	DB_TOKEN entry;			/* -> selected entry */
};

struct db_rec_select_list {
	int count;
	struct db_rec_select_1 list[0];
};

/* List of children under a directory */
struct db_rec_child_1 {
	struct {
		unsigned int primary: 1;
	} flags;
	int index;
	DB_TOKEN relname;
	DB_TOKEN entry;			/* -> source, target, directory */
};

struct db_rec_child_list {
	int count;
	struct {
		unsigned int sorted: 1;
	} flags;
	struct db_rec_child_1 list[0];
};

/* List of targets */
struct db_rec_target_1 {
	DB_TOKEN entry;			/* -> target */
};

struct db_rec_target_list {
	int count;
	struct db_rec_target_1 list[0];
};

/* List of components, configs or dependencies */
struct db_rec_ccd_1 {
	DB_TOKEN entry;			/* -> component, config or dependency entry */
	int index;			/* index of this entry when list was built */
	time_t mtime;			/* mtime of this entry when list was built */
};

struct db_rec_ccd_list {
	int count;
	int parent_index;		/* index of parent when list was built */
	time_t parent_mtime;		/* mtime of parent when list was built */
	struct {
		unsigned int sorted: 1;
	} flags;
	struct db_rec_ccd_1 list[0];
};

#define declare_db_list_size(U, l) \
	static inline int DB_SIZE_##U##_1(const DB_REC_##U##_1 *rec) { return(sizeof(*rec)); } \
	static inline int DB_SIZE_##U##_LIST(const DB_REC_##U##_LIST *rec) { return(sizeof(*rec)+rec->count*sizeof(rec->list[0])); }

declare_db_list_size(BASE, base)
declare_db_list_size(BASE_SELECT, base_select)
declare_db_list_size(SELECT, select)
declare_db_list_size(CHILD, child)
declare_db_list_size(TARGET, target)
declare_db_list_size(CCD, ccd)

extern const DB_TOKEN key_base;

/* ---- Functions ---- */

/* Read using a fixed size key */

extern DB_REC_FILENAME_LIST *read_db_filename_list(PPDB * db,
						   const char *filename,
						   int create);

extern DB_TOKEN new_db_directory(PPDB * db, const char *name);
extern DB_TOKEN findnew_db_directory(PPDB * db, const char *name);
extern DB_REC_DIRECTORY *read_db_directory(PPDB * db, DB_TOKEN token);
extern int sort_db_directories(PPDB * db);

extern DB_REC_SOURCE *read_db_source(PPDB * db, DB_TOKEN token);
extern DB_TOKEN new_db_source(PPDB * db, const char *name, int index);
extern DB_TOKEN findnew_db_source(PPDB * db, const char *name, int index);

extern DB_TOKEN new_db_target(PPDB * db, const char *name);
extern DB_TOKEN findnew_db_target(PPDB * db, const char *name);
extern DB_REC_TARGET *read_db_target(PPDB * db, DB_TOKEN token);

extern DB_TOKEN new_db_text(PPDB * db, const char *text, DB_TOKEN * token_p);
extern char *read_db_text(PPDB * db, DB_TOKEN token);
extern void update_db_text(PPDB * db, const char *text, DB_TOKEN token);
extern void empty_db_text(PPDB * db, DB_TOKEN token);

extern DB_TOKEN lookup_db_byname(PPDB * db, const char *name, DB_TYPE type,
				 int index);
extern DB_TOKEN lookup_db_byname_highest(PPDB * db, const char *name,
					 DB_TYPE type);
#define lookup_db_directory(db, name) lookup_db_byname(db, name, DB_RECTYPE_DIRECTORY, objtree_treename_index)
#define lookup_db_target(db, name) lookup_db_byname(db, name, DB_RECTYPE_TARGET, objtree_treename_index)

extern int run_db(PPDB * db, int (*userfunc) (PPDB *, kvpair, void *),
		  void *userdata);
extern int run_db_child(PPDB * db,
			int (*userfunc) (PPDB *, DB_REC_CHILD_1 *, void *),
			void *userdata, const char *relname);
extern int run_db_child_depth;
extern int run_db_select(PPDB * db,
			 int (*userfunc) (PPDB *, DB_REC_BASE_1 *,
					  DB_REC_SELECT_1 *, void *),
			 void *userdata, DB_REC_BASE_1 *);
extern int run_db_select_depth;

extern void unlink_db_files(PPDB * db, const char *relname);

extern void copy_db_data(PPDB * db, DB_TOKEN to, DB_TOKEN from);

extern DB_REC_BASE_1 *scan_db_base(PPDB * db, const char *name);
extern void add_db_base(PPDB * db, const char *name, int has_modules, int has_arch_head, int is_host, DB_TOKEN topdir);

extern PPDB *open_db_main(int open_type);
extern void close_db_main(PPDB * db);

extern DB_REC_BASE_LIST *rec_base;

#ifdef NEED_PPDB_INTERNALS

/* way oversized, 2.4.18 has ~40,000 fixed keys */
#define	ppdb_fixed_key_max		(512*1024)

/* 2**n fixed keys/values in each slab with a vector to point to each slab */
#define	ppdb_slab_shift			12
#define	ppdb_slab_num			(1 << ppdb_slab_shift)
#define	ppdb_fixed_key_vector_size	((ppdb_fixed_key_max + ppdb_slab_num - 1) / ppdb_slab_num)

/* 99% of variable records fit in 64 bytes, but allow for huge commands */
#define ppdb_variable_shift_min		6
#define ppdb_variable_shift_max		18
#define	ppdb_variable_vector_size	(ppdb_variable_shift_max - ppdb_variable_shift_min + 1)

static inline int
ppdb_order(int size)
{
	int i, j;
	for (i = ppdb_variable_shift_min, j = 1 << ppdb_variable_shift_min;
	     size > j && i <= ppdb_variable_shift_max;
	     ++i, j <<= 1) {
	};
	return (i - ppdb_variable_shift_min);
}

/* Approx. 19,000 files and directories in a full kernel build, including
 * objects.  Use a hash size which is a prime number, with room for expansion.
 */
#define ppdb_hashsize 32749

extern datum ppdb_fetch(PPDB *db, datum key);
extern void ppdb_store(PPDB *db, datum key, datum val, unsigned int flags);
extern kvpair ppdb_first(PPDB * db);
extern kvpair ppdb_next(PPDB * db);

#define	PPDB_INSERT	1
#define	PPDB_REPLACE	2

#define read_db_rec(db, key) \
({ \
	datum __tv, __tk; \
	__tk.dsize = sizeof(key); \
	__tk.dptr = &key; \
	__tv = ppdb_fetch(db, __tk); \
	__tv; \
})

#ifdef NDEBUG
#define assert_key_type(key)
#else
#define assert_key_type(key) \
({	DB_TOKEN __token = *((DB_TOKEN *)&key); \
	assert( \
		__token.type == DB_RECTYPE_BASE_LIST || \
		__token.type == DB_RECTYPE_FILENAME_LIST || \
		__token.type == DB_RECTYPE_DIRECTORY || \
		__token.type == DB_RECTYPE_BASE_SELECT_LIST || \
		__token.type == DB_RECTYPE_SELECT_LIST || \
		__token.type == DB_RECTYPE_TARGET || \
		__token.type == DB_RECTYPE_SOURCE || \
		__token.type == DB_RECTYPE_CHILD_LIST || \
		__token.type == DB_RECTYPE_TARGET_LIST || \
		__token.type == DB_RECTYPE_CCD_LIST || \
		__token.type == DB_RECTYPE_TEXT \
	      ); \
})
#endif

/* Insert or replace variable sized data, takes a datum value, requires a
 * fixed size key.
 */

#define insert_replace_db_rec_variable(db, key, value, flags) \
({ \
	datum __tk; \
	__tk.dsize = sizeof(key); \
	__tk.dptr = &key; \
	assert_key_type(key); \
	ppdb_store(db, __tk, value, flags); \
})

#define insert_db_rec_variable(db, key, value) \
	insert_replace_db_rec_variable(db, key, value, PPDB_INSERT)

#define replace_db_rec_variable(db, key, value) \
	insert_replace_db_rec_variable(db, key, value, PPDB_REPLACE)

/* Insert or replace fixed sized data, takes a structure value, requires a
 * fixed size key.
 */

#define insert_replace_db_rec_fixed(db, key, value, flags) \
({ \
	datum __tk, __tv; \
	__tk.dsize = sizeof(key); \
	__tk.dptr = (char *) &key; \
	__tv.dsize = sizeof(value); \
	__tv.dptr = &value; \
	assert_key_type(key); \
	ppdb_store(db, __tk, __tv, flags); \
})

#define insert_db_rec_fixed(db, key, value) \
	insert_replace_db_rec_fixed(db, key, value, PPDB_INSERT)

#define replace_db_rec_fixed(db, key, value) \
	insert_replace_db_rec_fixed(db, key, value, PPDB_REPLACE)

struct ppdb_key_fixed {
	unsigned int voff;	/* offset from start of database to value */
	DB_TOKEN token;
};

/* klen/vlen are free chain links when a slot is free */

struct ppdb_key_variable {
	unsigned int order;	/* order of slot, must be first */
	unsigned int klen;	/* length of key, must be second */
	unsigned int collision;	/* next key on collision chain */
	unsigned int voff;
	char value[0];
};

struct ppdb_value_variable {
	unsigned int order;	/* order of slot, must be first */
	unsigned int vlen;	/* length of value, must be second */
	unsigned int kvoff;	/* link to voff field that points to this value */
	char value[0];
};

static inline
int ppdb_fixed_key(DB_TYPE c)
{
	return(c !=  DB_RECTYPE_FILENAME_LIST);
}

static inline
int ppdb_fixed_value(DB_TYPE c)
{
	return(c == DB_RECTYPE_TARGET ||
	       c == DB_RECTYPE_SOURCE ||
	       c == DB_RECTYPE_DIRECTORY);
}

static char *ppdb_name = ".tmp_db_main";

struct ppdb_header {

        /* Master data that defines the database.  The first set of fields are
         * critical, any change to a critical field deletes the old database and
         * forces a complete kernel rebuild.
	 */
	int version;			/* database version */
	int fixed_slab_num;		/* value of ppdb_slab_num */
	int fixed_key_vector_size;	/* value of ppdb_fixed_key_vector_size */
	int variable_shift_min;		/* value of ppdb_variable_shift_min */
	int variable_shift_max;		/* value of ppdb_variable_shift_max */
	int hashsize;			/* value of ppdb_hashsize */

	/* 'self' must be the first non-critical field */
	struct ppdb_header *self;	/* Where I was loaded last time */
	size_t size;			/* Database size */
	
	struct {
		unsigned int dirty: 1;	/* database is dirty, verify it */
	} flags;
	int next_recno;			/* next record number to assign */

	/* Current slab for source values */
	DB_REC_SOURCE *source_slab;
	int source_slab_free;

	/* Current slab for target values */
	DB_REC_TARGET *target_slab;
	int target_slab_free;

	/* Current slab for directory values */
	DB_REC_DIRECTORY *directory_slab;
	int directory_slab_free;

	/* Free space at end of database */
	char *free_start;
	int free_left;

	/* Top level lookup for fixed sized keys */
	struct ppdb_key_fixed *key_fixed[ppdb_fixed_key_vector_size];

	/* Base of free lists for variable sized key/data */
	unsigned int variable_free[ppdb_variable_vector_size];

	/* Variable key hash buckets */
	unsigned int hash[ppdb_hashsize];
};

#define ppdb_extend	1024*1024	/* Arbitrary size to extend database */

extern struct ppdb_header *rec_master;

struct ppdb {
	struct ppdb_header *header;
	int fd;
	int key_type;			/* 1 - fixed, 2 - variable */
	int key_no;			/* fixed - key number, variable - bucket number */
	struct ppdb_key_variable *var_key;	/* current variable key */
};

#define ppdb_offset(db,x) ((char *)(x) - (char *)(db->header))
#define ppdb_pointer(db,x) ((void *)((char *)(db->header) + (x)))

#endif	/* NEED_PPDB_INTERNALS */

#endif				/* _PP_DB_H */
